'''
Weight Agnostic Neural Networks

To find architectures that can achieve much higher than chance accuracy using random weights
Use GA (or other random tool) to generate several candidates. -> choose the best one -> Finetune the last layer by MSE | or finetune the last few layers by back-propogation
Many neuroevolution algorithms have been defined. One common distinction is between algorithms that evolve only the strength of the connection weights for a fixed network topology (sometimes called conventional neuroevolution), as opposed to those that evolve both the topology of the network and its weights (called TWEANNs, for Topology and Weight Evolving Artificial Neural Network algorithms).
Most neural networks use gradient descent rather than neuroevolution. However, around 2017 researchers at Uber stated they had found that simple structural neuroevolution algorithms were competitive with sophisticated modern industry-standard gradient-descent deep learning algorithms, in part because neuroevolution was found to be less likely to get stuck in dead ends.
In direct encoding schemes the genotype directly maps to the phenotype. That is, every neuron and connection in the neural network is specified directly and explicitly in the genotype. In contrast, in indirect encoding schemes the genotype specifies indirectly how that network should be generated
'''

class WANN(object):
    pass # Research in future


'''
This ELM implementation is based on https://github.com/otenim/Numpy-ELM: 

MIT License

Copyright (c) 2019 Otenim

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import os
import numpy as np
import h5py
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import GridSearchCV, train_test_split
from tensorflow.keras.utils import to_categorical
from sklearn.metrics import log_loss, accuracy_score, recall_score
from tensorflow.keras.datasets import mnist
from sklearn.datasets import load_iris
# from keras import losses

def _mean_squared_error(y_true, y_pred):
    return 0.5 * np.mean((y_true - y_pred)**2)

def _mean_absolute_error(y_true, y_pred):
    return np.mean(np.abs(y_true - y_pred))

def _sigmoid(x):
    return 1. / (1. + np.exp(-x))

def _identity(x):
    return x

def _relu(x):
    return np.maximum(0, x)

def softmax(x):
    c = np.max(x, axis=-1)
    upper = np.exp(x - c)
    lower = np.sum(upper, axis=-1)
    return upper / lower

class ELM(object):
    def __init__(
        self, n_input_nodes, n_hidden_nodes, n_output_nodes,
        activation='sigmoid', loss='mean_squared_error', name=None,
        beta_init=None, alpha_init=None, bias_init=None):

        self.name = name
        self.__n_input_nodes = n_input_nodes
        self.__n_hidden_nodes = n_hidden_nodes
        self.__n_output_nodes = n_output_nodes

        # initialize weights and a bias
        if isinstance(beta_init, np.ndarray):
            if beta_init.shape != (self.__n_hidden_nodes, self.__n_output_nodes):
                raise ValueError(
                    'the shape of beta_init is expected to be (%d,%d).' % (self.__n_hidden_nodes, self.__n_output_nodes)
                )
            self.__beta = beta_init
        else:
            self.__beta = np.random.uniform(-1.,1.,size=(self.__n_hidden_nodes, self.__n_output_nodes))
        if isinstance(alpha_init, np.ndarray):
            if alpha_init.shape != (self.__n_input_nodes, self.__n_hidden_nodes):
                raise ValueError(
                    'the shape of alpha_init is expected to be (%d,%d).' % (self.__n_hidden_nodes, self.__n_output_nodes)
                )
            self.__alpha = alpha_init
        else:
            self.__alpha = np.random.uniform(-1.,1.,size=(self.__n_input_nodes, self.__n_hidden_nodes))
        if isinstance(bias_init, np.ndarray):
            if bias_init.shape != (self.__n_hidden_nodes,):
                raise ValueError(
                    'the shape of bias_init is expected to be (%d,).' % (self.__n_hidden_nodes,)
                )
            self.__bias = bias_init
        else:
            self.__bias = np.zeros(shape=(self.__n_hidden_nodes,))

        # set an activation function
        self.__activation = self.__get_activation_function(activation)

        # set a loss function
        self.__loss = self.__get_loss_function(loss)

    # FP
    def __call__(self, x):
        h = self.__activation(x.dot(self.__alpha) + self.__bias)
        return h.dot(self.__beta)

    def predict(self, x):
        return np.array(self(x))

    def evaluate(self, x, t, metrics=['loss']):
        y_pred = self.predict(x)
        y_true = t
        y_pred_argmax = np.argmax(y_pred, axis=-1)
        y_true_argmax = np.argmax(y_true, axis=-1)
        ret = []
        for m in metrics:
            if m == 'loss':
                loss = self.__loss(y_true, y_pred)
                ret.append(loss)
            elif m == 'accuracy':
                acc = np.mean(y_pred_argmax == y_true_argmax)
                ret.append(acc)
            elif m == 'precision':
                num_classes = t.shape[1] # len(t[0])
                precision = []
                for i in range(num_classes):
                    tp = np.sum((y_pred_argmax == i) & (y_true_argmax == i))
                    tp_fp = np.sum(y_pred_argmax == i)
                    precision.append(tp / tp_fp)
                ret.append(np.mean(precision))
            elif m == 'recall':
                num_classes = t.shape[1] # len(t[0])
                recall = []
                for i in range(num_classes):
                    tp = np.sum((y_pred_argmax == i) & (y_true_argmax == i))
                    tp_fn = np.sum(y_true_argmax == i)
                    recall.append(tp / tp_fn)
                ret.append(np.mean(recall))
            else:
                raise ValueError(
                    'an unknown evaluation indicator \'%s\'.' % m
                )
        if len(ret) == 1:
            ret = ret[0]
        elif len(ret) == 0:
            ret = None
        return ret


    def fit(self, x, t):
        H = self.__activation(x.dot(self.__alpha) + self.__bias)

        # compute a pseudoinverse of H
        H_pinv = np.linalg.pinv(H)

        # update beta
        self.__beta = H_pinv.dot(t)

    def save(self, filepath):
        with h5py.File(filepath, 'w') as f:
            arc = f.create_dataset('architecture', data=np.array([self.__n_input_nodes, self.__n_hidden_nodes, self.__n_output_nodes]))
            arc.attrs['activation'] = self.__get_activation_name(self.__activation).encode('utf-8')
            arc.attrs['loss'] = self.__get_loss_name(self.__loss).encode('utf-8')
            arc.attrs['name'] = self.name.encode('utf-8')
            f.create_group('weights')
            f.create_dataset('weights/alpha', data=self.__alpha)
            f.create_dataset('weights/beta', data=self.__beta)
            f.create_dataset('weights/bias', data=self.__bias)

    def __get_activation_function(self, name):
        if name == 'sigmoid':
            return _sigmoid
        elif name == 'identity':
            return _identity
        elif name == 'relu':
            return _relu
        else:
            raise ValueError(
                'an unknown activation function \'%s\'.' % name
            )

    def __get_activation_name(self, activation):
        if activation == _sigmoid:
            return 'sigmoid'
        elif activation == _identity:
            return 'identity'

    def __get_loss_function(self, name):
        if name == 'mean_squared_error':
            return _mean_squared_error
        elif name == 'mean_absolute_error':
            return _mean_absolute_error
        else:
            raise ValueError(
                'an unknown loss function \'%s\'.' % name
            )

    def __get_loss_name(self, loss):
        if loss == _mean_squared_error:
            return 'mean_squared_error'
        elif loss == _mean_absolute_error:
            return 'mean_absolute_error'
    
    @property
    def weights(self):
        return {
            'alpha': self.__alpha,
            'beta': self.__beta,
            'bias': self.__bias,
        }

    @property
    def input_shape(self):
        return (self.__n_input_nodes,)

    @property
    def output_shape(self):
        return (self.__n_output_nodes,)

    @property
    def n_input_nodes(self):
        return self.__n_input_nodes

    @property
    def n_hidden_nodes(self):
        return self.__n_hidden_nodes

    @property
    def n_output_nodes(self):
        return self.__n_output_nodes

    @property
    def activation(self):
        return self.__get_activation_name(self.__activation)

    @property
    def loss(self):
        return self.__get_loss_name(self.__loss)

def load_model(filepath):
    with h5py.File(filepath, 'r') as f:
        alpha_init = f['weights/alpha'][...]
        beta_init = f['weights/beta'][...]
        bias_init = f['weights/bias'][...]
        arc = f['architecture']
        n_input_nodes = arc[0]
        n_hidden_nodes = arc[1]
        n_output_nodes = arc[2]
        activation = arc.attrs['activation'].decode('utf-8')
        loss = arc.attrs['loss'].decode('utf-8')
        name = arc.attrs['name'].decode('utf-8')
        model = ELM(
            n_input_nodes=n_input_nodes,
            n_hidden_nodes=n_hidden_nodes,
            n_output_nodes=n_output_nodes,
            activation=activation,
            loss=loss,
            alpha_init=alpha_init,
            beta_init=beta_init,
            bias_init=bias_init,
            name=name,
        )
    return model


class ELMClassifierCV():

    def __init__(self, hparams = {'n_hidden_nodes': [1, 10], 'activation': ['relu', 'sigmoid', 'identity'] }):        
        '''
        parameters  
        ----------
        hparams : hyper-parameter candidate values to be grid-searched
        '''
        self.parameters =hparams
       
    def fit(self, X, y):
        elmc = ELMClassifier()
        self.clf = GridSearchCV(elmc, self.parameters, scoring = 'accuracy') # 'neg_log_loss'
        self.clf.fit(X, y)
        # print( sorted(clf.cv_results_.keys()) )
        return self

    def predict_proba(self, X):
        return self.clf.predict_proba(X)

    def predict(self, X):
        return self.clf.predict(X)


class ELMClassifier(BaseEstimator, ClassifierMixin):
    '''
    A wrapper for ELM.
    Encapsulate as a sklearn-compatible base learner/estimator
    '''
    
    def __init__(self, n_hidden_nodes = 20, activation = 'relu'):
        
        # ===============================
        # ELM parameters
        # ===============================
        self.n_hidden_nodes = n_hidden_nodes #x_train.shape[1]
        self.loss = 'mean_squared_error' # 'mean_absolute_error'
        self.activation = activation # 'sigmoid' # 'identity' # 'relu'
       
    def fit(self, X, y):
        
        # ===============================
        # Instantiate ELM
        # ===============================
        if len(y.shape) == 2: #one-hot
            n_classes = y.shape[1]
        else:
            n_classes = len(set(y))

        self.model = ELM(
            n_input_nodes = X.shape[1],
            n_hidden_nodes = self.n_hidden_nodes,
            n_output_nodes = n_classes,
            loss = self.loss,
            activation = self.activation,
            name = 'elm'
        )
        
        self.model.fit(X, y)

    def predict_proba(self, X):        
        yh = self.model.predict(X)
        # print(yh) # array e.g., 0.37854525 0.         0.         
        return yh

    def predict(self, X):
        yh = self.model.predict(X) # (m,  n_classes)
        if (len(yh.shape) <= 1): # e.g., (m, )
            return yh > 0.5 # default 0.5 threshold
        yh = np.argmax(yh, axis=-1)
        # print(yh.shape)
        return yh

    def evaluate(self, val_x, val_y, metrics=['loss', 'accuracy']):

        if len(val_y.shape) == 2: #one-hot
            n_classes = val_y.shape[1]
            y_gt = val_y
            val_y = val_y.argmax(-1)
        else:
            n_classes = len(set(val_y))
            y_gt = to_categorical(val_y, len(set(val_y))).astype(np.float32)
        
        y_hat = self.predict_proba(val_x)
        y_pred = self.predict(val_x)
        # print(val_y.shape, y_gt.shape, y_hat.shape, y_pred.shape)
        
        val_loss = log_loss(y_gt, y_hat)
        val_acc = accuracy_score(val_y, y_pred)
        val_recall = recall_score(val_y, y_pred, average = 'macro')
        
        return val_loss, val_acc

    def save_model(self, path):
        self.model.save(path)

    def load_model(self, path):
        self.model = load_model(path)

    def run_example(x_train, x_test, t_train, t_test, save_model_path = ''):
        
        # ===============================
        # ELM parameters
        # ===============================
        n_hidden_nodes = x_train.shape[1]
        loss = 'mean_squared_error' # 'mean_absolute_error'
        activation = 'sigmoid' # 'identity'
        n_classes = t_train.shape[1]

        # ===============================
        # Instantiate ELM
        # ===============================
        clf = ELMClassifier()

        # ===============================
        # Training
        # ===============================
        clf.fit(x_train, t_train)
        train_loss, train_acc = clf.evaluate(x_train, t_train, metrics=['loss', 'accuracy'])
        print('train_loss: %f' % train_loss) # loss value
        print('train_acc: %f' % train_acc) # accuracy

        # ===============================
        # Validation
        # ===============================
        val_loss, val_acc = clf.evaluate(x_test, t_test, metrics=['loss', 'accuracy'])
        print('val_loss: %f' % val_loss)
        print('val_acc: %f' % val_acc)

        # ===============================
        # Prediction
        # ===============================
        print("\n\n========== prediction on the first 10 test samples ===========\n")
        x = x_test[:10]
        t = t_test[:10]
        y_pred = clf.predict_proba(x)

        for i in range(len(y_pred)):
            print('---------- prediction %d ----------' % (i+1))
            class_pred = np.argmax(y_pred[i])
            prob_pred = y_pred[i][class_pred]
            class_true = np.argmax(t[i])
            print('class: %d, probability: %f' % (class_pred, prob_pred))
            print('class (true): %d' % class_true)

        if (save_model_path):

            # ===============================
            # Save model
            # ===============================
            clf.save_model(save_model_path)

            # ===============================
            # Load model
            # ===============================
            # clf.load_model(save_model_path)

    def run_iris_example():

        # ===============================
        # Load dataset
        # ===============================
        iris = load_iris()
        n_classes = len(set(iris.target))
        # stdsc = StandardScaler()
        # irisx = stdsc.fit_transform(iris.data)
        x_train, x_test, t_train, t_test = train_test_split(iris.data, iris.target, test_size=0.2)
        t_train = to_categorical(t_train, n_classes).astype(np.float32)
        t_test = to_categorical(t_test, n_classes).astype(np.float32)

        ELMClassifier.run_example(x_train, x_test, t_train, t_test)


    def run_mnist_example():

        n_classes = 10
        (x_train, t_train), (x_test, t_test) = mnist.load_data()

        # ===============================
        # Preprocess
        # ===============================
        x_train = x_train.astype(np.float32) / 255.
        x_train = x_train.reshape(-1, 28**2)
        x_test = x_test.astype(np.float32) / 255.
        x_test = x_test.reshape(-1, 28**2)
        t_train = to_categorical(t_train, n_classes).astype(np.float32)
        t_test = to_categorical(t_test, n_classes).astype(np.float32)

        ELMClassifier.run_example(x_train, x_test, t_train, t_test)

def create_elm_instance(L, activation = 'relu'):
    '''
    Parameters
    ----------
    L : n_hidden_nodes
    activation : {'relu', 'sigmoid', 'identity'}
    '''
    return ELMClassifier(n_hidden_nodes = L, activation = activation) # 'identity'


def create_elmcv_instance():
    return ELMClassifierCV()

def ElmRegression(x_train, x_test, t_train, t_test, L = 10, rndb = False):
    
    # ===============================
    # ELM parameters
    # ===============================
    n_input_nodes = x_train.shape[1]
    n_hidden_nodes = max(x_train.shape[1], L) # at least 10 hidden nodes
    n_output_nodes = t_train.shape[1]

    # ===============================
    # Initialization
    # ===============================
    A = np.random.uniform(-1.,1.,size=(n_input_nodes, n_hidden_nodes))
    if rndb:
        b = np.random.uniform(-1.,1.,size=(n_hidden_nodes,))
    else:
        b = np.zeros(shape=(n_hidden_nodes,))

    # ===============================
    # Solve
    # ===============================
    H = _sigmoid(x_train @ A + b)
    B = np.linalg.pinv(H) @ t_train

    # ===============================
    # Evaluate on CV dataset
    # ===============================
    H = _sigmoid(x_test @ A + b)
    t_pred = H @ B

    MSE = np.mean((t_test - t_pred)**2)
    return MSE, A, b, B


from inspect import stack
from numpy.linalg import multi_dot
import numpy as np
from sklearn.preprocessing import OneHotEncoder
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.metrics import log_loss, accuracy_score, precision_score, recall_score
from tensorflow.keras.utils import to_categorical
from sklearn.model_selection import train_test_split, GridSearchCV

import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics.pairwise import cosine_similarity, rbf_kernel, linear_kernel, sigmoid_kernel, chi2_kernel, polynomial_kernel, \
                                        additive_chi2_kernel, laplacian_kernel
from sklearn.preprocessing import OneHotEncoder

class KNNRW(object):
    """
    kernel-NNRW
    """

    def __init__(self, N = 1, 
    type = 'classification', 
    kernels = ['linear','sigmoid'], 
    flavor = 'stack' 
    ):

        '''
        Parameters
        ----------
        N : int. Some kernels' params need n_features to get the default value.

        kernels : int or array. If kernels is an integer n, will use the first n built-in kernels. 
                  kernels can also be an array of kernel names (duplicate is allowed)

        kernels = ['linear','sigmoid'] is the default case: linear + sigmoid. It is kernel version of vanilla RVFL
        kernels = ['linear'] reduces to a linear model
        kernels = ['sigmoid'] reduceds to k-ELM.
        kernels > 2 will add more types of non-linear kernels

        flavor: 'stack' , 'sum', or 'stack+sum'.

        '''
        self.type = type
        self.flavor = flavor

        # the candidate kernels are pre-selected by Raman domain datasets
        self.kernel_dict = {"linear": lambda x, y = None: linear_kernel(x, y),
                            "sigmoid_0.1": lambda x, y = None: sigmoid_kernel(x, y, 0.1/N),
                            "sigmoid": lambda x, y = None: sigmoid_kernel(x, y, 1.0/N),
                            "sigmoid_10": lambda x, y = None: sigmoid_kernel(x, y, 10.0/N),
                            "sigmoid_100": lambda x, y = None: sigmoid_kernel(x, y, 100.0/N),
                            "rbf_0.1": lambda x, y = None: rbf_kernel(x, y, 0.1/N),
                            "rbf_0.3": lambda x, y = None: rbf_kernel(x, y, 0.3/N),
                            "rbf": lambda x, y = None: rbf_kernel(x, y),
                            "rbf_3": lambda x, y = None: rbf_kernel(x, y, 3.0/N),
                            "rbf_10": lambda x, y = None: rbf_kernel(x, y, 10.0/N),    
                            "laplace_0.5": lambda x, y = None: laplacian_kernel(x, y, 0.5/N),
                            "laplace": lambda x, y = None: laplacian_kernel(x, y),
                            "laplace_2": lambda x, y = None: laplacian_kernel(x, y, 2.0/N),
                            "laplace_4": lambda x, y = None: laplacian_kernel(x, y, 4.0/N),
                            "cosine": lambda x, y = None: cosine_similarity(x, y),                            
                            "chi2_0.001": lambda x, y = None: chi2_kernel(x, y, 0.001),
                            "chi2_0.01": lambda x, y = None: chi2_kernel(x, y, 0.01),
                            "chi2_0.1": lambda x, y = None: chi2_kernel(x, y, 0.1),
                            "chi2": lambda x, y = None: chi2_kernel(x, y),
                            "add_chi2": lambda x, y = None: additive_chi2_kernel(x, y),
                            "poly": lambda x, y = None: polynomial_kernel(x, y, 0.001/N)
                            }    

        self.kernel_names = list(self.kernel_dict.keys())

        if isinstance(kernels,int): # type(kernels) == 'int': # int
            # check whether kernel is in valid range
            if (kernels < 1):
                kernels = 1
            if (kernels > len(self.kernel_names ) ):
                kernels = len(self.kernel_names )

            self.kernels = []
            for i in range(kernels):
                self.kernels.append(self.kernel_names[i])

        else: # array
            self.kernels = kernels

    def _transform_label(self, y):
        enc = OneHotEncoder(handle_unknown='ignore')
        try:
            target = enc.fit_transform(y).toarray()
            # print('the label can be transformed directly using onehotencoder')
        except:
            target = enc.fit_transform(y.reshape(-1, 1)).toarray()
            # print('the label must be reshaped before being transformed')
        return target

    def _softmax(self, x):
        out = np.exp(x)
        return out/ np.sum(out, axis=1, keepdims=True)

    def fit(self, train_x, train_y, verbose = 0):
        """
        Params:
        ---------
        :param train_x: a NumofSamples * NumofFeatures matrix, training data
        :param train_y: training label
        """

        n_features = train_x.shape[1]

        summed_omegas = self.kernel_dict[self.kernels[0]](train_x) # linear kernel
        stacked_omega = self.kernel_dict[self.kernels[0]](train_x)

        # add up the other non-linear kernels 
        for i in range(1, len(self.kernels) ): 
            omega = self.kernel_dict[self.kernels[i]](train_x)
            summed_omegas = omega + summed_omegas
            stacked_omega = np.hstack((stacked_omega,omega))

        # omega1, omega2 = self.kernel_dict["linear"](train_x), \
        #                 self.kernel_dict[self.kernel_name](train_x) # linear + non-linear combination
        
        if self.type == 'classification':
            one_hot_target = self._transform_label(train_y)
        elif self.type == 'regression':
            one_hot_target = train_y
        else:
            raise Exception("The type is not supported now! please select classification or regression.")
        
        if self.flavor == 'stack':
            OMEGA = stacked_omega
        elif self.flavor == 'sum':
            OMEGA = summed_omegas
        else: # stack + sum
            OMEGA = np.hstack((stacked_omega, summed_omegas))

        self.beta = np.linalg.pinv(OMEGA) @ one_hot_target
        if verbose:
            print('OMEGA shape: ', OMEGA.shape, ', BETA shape: ', self.beta.shape)

    def predict(self, train_x, test_x):
        """
        :param train_x:  a NumofSamples * NumofFeatures matrix, training data, building the kernel matrix
        :param test_x: a NumofTestSamples * NumofFeatures matrix, test data
        :return: y_hat, the predicted labels
        """
        y_hat = self.predict_proba(train_x, test_x)
        if self.type == 'classification':
            y_hat = np.argmax(y_hat, axis=1)
            return y_hat
        else:
            return y_hat

    def predict_proba(self, train_x, test_x):

        # NOTE: Unlike training, in prediction we use (test_x, train_x) as kernel input. train_x acts as anchors of the kernel operations.
        summed_omegas = self.kernel_dict[self.kernels[0]](test_x, train_x) # linear kernel
        stacked_omega = self.kernel_dict[self.kernels[0]](test_x, train_x)

        # add up the other non-linear kernels 
        for i in range(1, len(self.kernels)) :
            omega = self.kernel_dict[self.kernels[i]](test_x, train_x)
            summed_omegas = omega + summed_omegas
            stacked_omega = np.hstack((stacked_omega,omega))

        if self.flavor == 'stack':
            OMEGA = stacked_omega
        elif self.flavor == 'sum':
            OMEGA = summed_omegas
        else: # stack + sum
            OMEGA = np.hstack((stacked_omega, summed_omegas))

        y_hat_temp = (OMEGA).dot(self.beta)
        if self.type == "classification":
            y_hat_prob = self._softmax(y_hat_temp)
            return y_hat_prob
        else:
            return y_hat_temp


def KNNRWGridSearch(x_train, x_test, t_train, t_test, verbose = 0):

    N = x_train.shape[1]
    K = 22

    taccs = []
    vaccs = []
        
    # Perform a grid search
    for flavor in ['sum','stack','stack+sum']:
        
        tacc = []
        vacc = []
        kernels = []
        
        for k in list(range(1, K)):
            
            # ===============================
            # Instantiate
            # ===============================
            clf = KNNRWClassifier(N = N, kernels = k, flavor = flavor)

            if verbose:
                print('===== Flavor: ' + flavor + ', kernel = ' + str(clf.model.kernels) + ' =====')
            kernels.append( str(clf.model.kernels) )

            # ===============================
            # Training
            # ===============================
            clf.fit(x_train, t_train, verbose = verbose)
            t_pred = clf.predict(x_train, x_train)
            ACC = np.mean(t_train == t_pred)
            if verbose:
                print("ACC on training set = ", ACC)
            tacc.append(ACC)

            # ===============================
            # Validation
            # ===============================
            t_pred = clf.predict(x_train, x_test)
            ACC = np.mean(t_test == t_pred)
            if verbose:
                print("ACC on test set = ", ACC)
            vacc.append(ACC)
            
            if verbose:
                print('========= \n\n')
            
        taccs.append(tacc)
        vaccs.append(vacc)

    for i, flavor in enumerate(['sum','stack','stack+sum']):
        plt.figure(figsize = (K, 4))
        plt.title(flavor)
        plt.scatter(kernels, taccs[i], label = 'train acc')
        plt.scatter(kernels, vaccs[i], label = 'val acc')
        plt.legend()
        plt.xticks(rotation=-90)
        plt.show()

    return taccs, vaccs    


class KNNRWClassifier(BaseEstimator, ClassifierMixin):
    '''
    Encapsulate as a sklearn estimator
    '''
    def __init__(self, N = 1, kernels = 5, flavor = 'stack'):
        self.model = KNNRW(N = N, kernels = kernels, type="classification", flavor=flavor) 

    def fit(self, X, y, verbose = 0):     
        
        self.model.fit(X, y, verbose = verbose)
        self.classes_ = np.array(list(set(y)))

    def predict(self, X_train, X_test):
        return self.model.predict(X_train, X_test)

    def predict_proba(self, X_train, X_test):
        return self.model.predict_proba(X_train, X_test)    

def create_knnrw_instance(L): # L is the non-linear kernel number
    return KNNRWClassifier(L)


'''
Radial basis function network

A RBF SVM would be virtually equivalent to a RBF neural nets where the weights of the first layer would be fixed to the feature values of all the training samples. Only the second layer weights are tuned by the learning algorithm. This allows the optimization problem to be convex hence admit a single global solution. The fact that the number of the potential hidden nodes can grow with the number of samples makes this hypothetical neural network a non-parametric model (which is usually not the case when we train neural nets: we tend to fix the architecture in advance as a fixed hyper-parameter of the algorithm independently of the number of samples).
    RBF neural nets have a higher number of hyper-parameters (the bandwidth of the RBF kernel, number of hidden nodes + the initialization scheme of the weights + the strengths of the regularizer a.k.a weight decay for the first and second layers + learning rate + momentum) + the local optima convergence issues (that may or not be an issue in practice depending on the data and the hyper-parameter)
    RBF SVM has 2 hyper-parameters to grid search (the bandwidth of the RBF kernel and the strength of the regularizer) and the convergence is independent from the init (convex objective function)

A rbf implementation in keras
ref: https://github.com/PetraVidnerova/rbf_keras

𝑅𝐵𝐹(𝑥,𝑥′)=𝑒 (−||𝑥−𝑥′|| / 2𝜎^2 )

参数beta： 𝛽=1 / 2𝜎^2

'''

from tensorflow.keras import backend as K
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Layer, Dense
from tensorflow.keras.optimizers import RMSprop
from tensorflow.keras.initializers import RandomUniform, Initializer, Constant

import math
import numpy as np
from sklearn.cluster import KMeans

class InitCentersKMeans(Initializer):
    """ Initializer for initialization of centers of RBF network
        by clustering the given data set.
    # Arguments
        X: matrix, dataset
    """

    def __init__(self, X, max_iter=100):
        self.X = X
        self.max_iter = max_iter

    def __call__(self, shape, dtype=None):
        assert shape[1] == self.X.shape[1]

        n_centers = shape[0]
        km = KMeans(n_clusters=n_centers, max_iter=self.max_iter, verbose=0)
        km.fit(self.X)
        return km.cluster_centers_


class InitCentersRandom(Initializer):
    """ Initializer for initialization of centers of RBF network
        as random samples from the given data set.
    # Arguments
        X: matrix, dataset to choose the centers from (random rows
          are taken as centers)
    """

    def __init__(self, X):
        self.X = X

    def __call__(self, shape, dtype=None):
        assert shape[1] == self.X.shape[1]
        idx = np.random.randint(self.X.shape[0], size=shape[0])
        return self.X[idx, :]


class RBFLayer(Layer):
    """ Layer of Gaussian RBF units.
    # Example
    ```python
        model = Sequential()
        model.add(RBFLayer(10,
                           initializer=InitCentersRandom(X),
                           betas=1.0,
                           input_shape=(1,)))
        model.add(Dense(1))
    ```
    # Arguments
        output_dim: number of hidden units (i.e. number of outputs of the
                    layer)
        initializer: instance of initiliazer to initialize centers
        betas: float, initial value for betas
    """

    def __init__(self, output_dim, initializer=None, betas=1.0, **kwargs):
        self.output_dim = output_dim
        self.init_betas = betas
        if not initializer:
            self.initializer = RandomUniform(0.0, 1.0)
        else:
            self.initializer = initializer
        super(RBFLayer, self).__init__(**kwargs)

    def build(self, input_shape):

        self.centers = self.add_weight(name='centers',
                                       shape=(self.output_dim, input_shape[1]),
                                       initializer=self.initializer,
                                       trainable=True)
        self.betas = self.add_weight(name='betas',
                                     shape=(self.output_dim,),
                                     initializer=Constant(
                                         value=self.init_betas),
                                     # initializer='ones',
                                     trainable=True)

        super(RBFLayer, self).build(input_shape)

    def call(self, x):

        C = K.expand_dims(self.centers)
        H = K.transpose(C-K.transpose(x))
        return K.exp(-self.betas * K.sum(H**2, axis=1))

        # C = self.centers[np.newaxis, :, :]
        # X = x[:, np.newaxis, :]

        # diffnorm = K.sum((C-X)**2, axis=-1)
        # ret = K.exp( - self.betas * diffnorm)
        # return ret

    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.output_dim)

    def get_config(self):
        # have to define get_config to be able to use model_from_json
        config = {
            'output_dim': self.output_dim
        }
        base_config = super(RBFLayer, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


from sklearn.base import BaseEstimator, ClassifierMixin

class RBFNN(BaseEstimator, ClassifierMixin):
    '''
    A wrapper for RBF-NN.
    Encapsulate as a sklearn-compatible base learner/estimator
    '''
    
    def __init__(self, n_hidden_nodes = 20, flavor = 'classifier'):
        
        # ===============================
        # parameters
        # ===============================
        self.n_hidden_nodes = n_hidden_nodes # RBF node numbers
        self.flavor = flavor # 'classifier' # or 'regressor'

    def fit(self, X, y, batch_size = 16, epochs = 20):
        
        # ===============================
        # Instantiate
        # ===============================
        self.model = Sequential()
        self.rbflayer = RBFLayer(output_dim = self.n_hidden_nodes,
                            initializer = InitCentersRandom(X), # InitCentersKMeans(X)
                            betas=2.0,
                            input_shape=(X.shape[1],))
        self.model.add(self.rbflayer)
        if self.flavor == 'classifier' and len(y.shape) == 2: # y should be one-hot 
            self.model.add(Dense(y.shape[1]))
            num_classes = y.shape[1]
        else:
            num_classes = 2
            self.model.add(Dense(1)) # binary classification or regression

        if self.flavor == 'classifier':
            if num_classes == 2:
                self.model.compile(loss='binary_crossentropy', optimizer=RMSprop())
            else:
                self.model.compile(loss='categorical_crossentropy', optimizer=RMSprop())
        else:
            self.model.compile(loss='mean_squared_error', optimizer=RMSprop())
        
        self.model.fit(X, y, batch_size=batch_size,
          epochs=epochs)

        self.rbf_weights = self.rbflayer.weights
        self.rbf_centers = self.rbflayer.get_weights()[0] # centroids, landmarks
        self.rbf_betas = self.rbflayer.get_weights()[1].tolist()
        self.rbf_sigmas = list( map(lambda x: math.sqrt(x/2.0), self.rbf_betas) )

    #def predict_proba(self, X):        
    #    yh = self.model.predict(X)
    #    # print(yh) # array e.g., 0.37854525 0.         0.         
    #    return yh

    def predict(self, X):
        yh = self.model.predict(X)
        if (len(yh.shape) <= 1):
            return yh > 0.5 # default 0.5 threshold
        yh = np.argmax(yh, axis=-1)
        # print(yh.shape)
        return yh

class RBFNNClassifier(RBFNN):

    def isRBFNN(self):
        return True

    def __init__(self, n_hidden_nodes = 20):

        # invoking the __init__ of the parent class
        RBFNN.__init__(self, n_hidden_nodes, 'classfier')   
        

def create_rbfnn_instance(L, activation = 'relu'):
    '''
    Parameters
    ----------
    L : n_hidden_nodes
    '''
    return RBFNNClassifier(n_hidden_nodes = L)


'''
Auto Encoder with Random Weights

H_Encoder =>   
latent z (how to decide dim and target values) =>  
H_Decoder

The encoder could be a RP (random projection) DR layer. 
'''

class AEWW(object):
    pass # Research in future


from sklearn.linear_model import LogisticRegression, LogisticRegressionCV

def create_lr_instance(C = 1.0, penalty = 'l2', multi_class = 'auto'):

    '''
    Create an sklearn.linear_model.LogisticRegression instance.

    Parameters
    ----------
    penalty : {‘l1’, ‘l2’, ‘elasticnet’, ‘none’}
    C : Inverse of regularization strength
    multi_class : {‘auto’, ‘ovr’, ‘multinomial’}
    '''
    return LogisticRegression(C = C, penalty = penalty, multi_class = multi_class)


from .elm import *
from .rvfl import *
from .mlp import *
from .dtc import *
from .lr import *
from .knn import *
import matplotlib.pyplot as plt
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import train_test_split, GridSearchCV
from sklearn.ensemble import StackingClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.svm import SVC
from sklearn.multiclass import OneVsOneClassifier, OneVsRestClassifier
from keras.utils import to_categorical
import numpy as np
from tqdm import tqdm
import time
import matplotlib.ticker as mticker

import warnings
warnings.filterwarnings("ignore")

#region Performance Test. Compare NNRW with mainstream classifier models.

def ELMClf(X, y, L = 100, verbose = False):
    '''
    A wrapper for ELM.
    L represents the number of hidden layer nodes.
    '''

    # ===============================
    # Load dataset
    # ===============================
    n_classes = len(set(y))
    x_train, x_test, t_train, t_test = train_test_split(X, y, test_size=0.2)
    t_train = to_categorical(t_train, n_classes).astype(np.float32)
    t_test = to_categorical(t_test, n_classes).astype(np.float32)
    # print(x_train.shape, x_test.shape, t_train.shape, t_test.shape)

    # ===============================
    # ELM parameters
    # ===============================
    n_hidden_nodes = L #x_train.shape[1]
    loss = 'mean_squared_error' # 'mean_absolute_error'
    activation = 'sigmoid' # 'identity'

    # ===============================
    # Instantiate ELM
    # ===============================
    model = ELM(
        n_input_nodes = x_train.shape[1],
        n_hidden_nodes = n_hidden_nodes,
        n_output_nodes = n_classes,
        loss = loss,
        activation=activation,
        name='elm'
    )

    # ===============================
    # Training
    # ===============================
    time1 = time.time_ns()        
    model.fit(x_train, t_train)
    time2 = time.time_ns()
    # print('ELM took {:.3f} ms'.format((time2-time1)/ (10 ** 6)))
    
    train_loss, train_acc, train_precision, train_recall = model.evaluate(x_train, t_train, 
    metrics=['loss', 'accuracy', 'precision', 'recall'])
    if (verbose):
        print('train_loss: %f' % train_loss) # loss value
        print('train_acc: %f' % train_acc) # accuracy
        print('train_precision: %f' % train_precision)
        print('train_recall: %f' % train_recall) # uar (unweighted average recall)


    # ===============================
    # Validation
    # ===============================
    val_loss, val_acc, val_precision, val_recall = model.evaluate(x_test, t_test, 
    metrics=['loss', 'accuracy', 'precision', 'recall'])
    if (verbose):
        print('val_loss: %f' % val_loss)
        print('val_acc: %f' % val_acc)
        print('val_precision: %f' % val_precision)
        print('val_recall: %f' % val_recall)
        print('train/fit time: {:.3f} ms'.format((time2-time1)/ (10 ** 6)))
    
    return train_acc, val_acc, (time2-time1)/ (10 ** 6)


def RVFLClf(X, y, L = 100, verbose = False):
    '''
    A wrapper for RVFL.
    L represents the number of hidden layer nodes.
    '''

    # ===============================
    # Load dataset
    # ===============================
    n_classes = len(set(y))
    x_train, x_test, t_train, t_test = train_test_split(X, y, test_size=0.2)
    
    ##  NOTE: Due to the inner implementation, RVFL doesn't need to do this one-hot encoding.
    # t_train = to_categorical(t_train, n_classes).astype(np.float32)
    # t_test = to_categorical(t_test, n_classes).astype(np.float32)
    
    # ===============================
    # RVFL parameters
    # ===============================
    n_hidden_nodes = L #x_train.shape[1]
    loss = 'mean_squared_error' # 'mean_absolute_error'
    activation = 'sigmoid' # 'identity'

    # ===============================
    # Instantiate RVFL
    # ===============================
    model = RVFL(
        n_hidden_nodes,
        random_type="uniform",
        type="classification"
    )

    # ===============================
    # Training
    # ===============================
    time1 = time.time_ns()        
    model.fit(x_train, t_train)
    time2 = time.time_ns()
    # print('RVFL took {:.3f} ms'.format((time2-time1)/ (10 ** 6)))
        
    train_loss, train_acc = model.evaluate(x_train, t_train)
    if (verbose):
        print('train_loss: %f' % train_loss) # loss value
        print('train_acc: %f' % train_acc) # accuracy
        print('train_precision: %f' % train_precision)
        print('train_recall: %f' % train_recall) # uar (unweighted average recall)


    # ===============================
    # Validation
    # ===============================
    val_loss, val_acc = model.evaluate(x_test, t_test)
    if (verbose):
        print('val_loss: %f' % val_loss)
        print('val_acc: %f' % val_acc)
        print('val_precision: %f' % val_precision)
        print('val_recall: %f' % val_recall)
        print('train/fit time: {:.3f} ms'.format((time2-time1)/ (10 ** 6)))
    
    return train_acc, val_acc, (time2-time1)/ (10 ** 6)


def MLPClf(X, y, L = 100, verbose = False, alpha = 0.1, iters = 150):
    '''
    A wrapper for MLP, which implements a GD (gradient-descent) training.
    L represents the number of hidden layer nodes.
    '''

    # ===============================
    # Load dataset
    # ===============================
    n_classes = len(set(y))
    x_train, x_test, t_train, t_test = train_test_split(X, y, test_size=0.2)
    t_train = to_categorical(t_train, n_classes).astype(np.float32)
    t_test = to_categorical(t_test, n_classes).astype(np.float32)
    
    # ===============================
    # Instantiate MLP
    # ===============================
    network = MLP(input_size=X.shape[1], hidden_size=L, output_size = n_classes)

    # ===============================
    # Training - MBGD
    # ===============================
    time1 = time.time_ns()       
    
    train_size = x_train.shape[0]
    batch_size = round(train_size / 2) 
    learning_rate = alpha

    iter_per_epoch = max(train_size / batch_size, 1)

    for i in range(iters): # 学习率=0.1时，100是最小可接受的值，低于此值，acc会变低，并出现较大波动
        batch_mask = np.random.choice(train_size, batch_size)
        x_batch = x_train[batch_mask]
        t_batch = t_train[batch_mask]

        # grad = network.numerical_gradient(x_batch, t_batch)
        grad = network.bp_gradient(x_batch, t_batch) # BP is much faster than the numerical gradient method

        # update parameters
        for key in ('W1', 'b1', 'W2', 'b2'):
            network.params[key] -= learning_rate * grad[key]

    time2 = time.time_ns()
        
    train_acc = network.accuracy(x_train, t_train)
    if (verbose):
        print('train_acc: %f' % train_acc) # accuracy

    # ===============================
    # Validation
    # ===============================
    val_acc = network.accuracy(x_test, t_test)
    if (verbose):
        print('val_acc: %f' % val_acc)
        print('train/fit time: {:.3f} ms'.format((time2-time1)/ (10 ** 6)))
    
    return train_acc, val_acc, (time2-time1)/ (10 ** 6)


def SVMClf(X, y, gamma = 1, verbose = False):
    '''
    A wrapper for SVM. Internally, we use the one-vs-one strategy for multi-class.
    gamma represents the gamma hyper-parameter of the rbf kernel, i.e., 1/ (2*sigma^2).
    '''

    # ===============================
    # Load dataset
    # ===============================
    n_classes = len(set(y))
    x_train, x_test, t_train, t_test = train_test_split(X, y, test_size=0.2)
    #t_train = to_categorical(t_train, n_classes).astype(np.float32)
    #t_test = to_categorical(t_test, n_classes).astype(np.float32)
    # print(x_train.shape, x_test.shape, t_train.shape, t_test.shape)

    # ===============================
    # Instantiate SVC
    # ===============================
    
    model = OneVsOneClassifier( SVC(kernel='rbf', gamma = gamma, C = 2) ) # C = 2 is selected by CV

    # ===============================
    # Training
    # ===============================
    time1 = time.time_ns()        
    model.fit(x_train, t_train)
    time2 = time.time_ns()
    # print('ELM took {:.3f} ms'.format((time2-time1)/ (10 ** 6)))
        
        
    t_pred = model.predict(x_train)
    train_acc = np.mean(t_train == t_pred)

    # train_loss, train_acc, train_precision, train_recall = model.evaluate(x_train, t_train, metrics=['loss', 'accuracy', 'precision', 'recall'])
    if (verbose):
        print('train_acc: %f' % train_acc) # accuracy
        #print('train_precision: %f' % train_precision)
        #print('train_recall: %f' % train_recall) # uar (unweighted average recall)


    # ===============================
    # Validation
    # ===============================
    t_pred = model.predict(x_test)
    val_acc = np.mean(t_test == t_pred)
    
    # val_loss, val_acc, val_precision, val_recall = model.evaluate(x_test, t_test, metrics=['loss', 'accuracy','precision', 'recall'])
    if (verbose):
        print('val_acc: %f' % val_acc)
        #print('val_precision: %f' % val_precision)
        #print('val_recall: %f' % val_recall)
        print('train/fit time: {:.3f} ms'.format((time2-time1)/ (10 ** 6)))
    
    return train_acc, val_acc, (time2-time1)/ (10 ** 6)


def TreeClf(X, y, max_depth = 1, verbose = False):
    '''
    A wrapper for DTC (decision tree classifier). Internally, we use the one-vs-one strategy for multi-class.
    max_depth represents the maximum tree depth.
    '''

    # ===============================
    # Load dataset
    # ===============================
    n_classes = len(set(y))
    x_train, x_test, t_train, t_test = train_test_split(X, y, test_size=0.2)

    # ===============================
    # Instantiate DTC
    # ===============================
    
    model =  DecisionTreeClassifier(criterion = 'entropy', max_depth = max_depth)  # A 1-deep tree is also called a decision stump

    # ===============================
    # Training
    # ===============================
    time1 = time.time_ns()        
    model.fit(x_train, t_train)
    time2 = time.time_ns()
    # print('ELM took {:.3f} ms'.format((time2-time1)/ (10 ** 6)))
        
        
    t_pred = model.predict(x_train)
    train_acc = np.mean(t_train == t_pred)

    # train_loss, train_acc, train_precision, train_recall = model.evaluate(x_train, t_train, metrics=['loss', 'accuracy', 'precision', 'recall'])
    if (verbose):
        print('train_acc: %f' % train_acc) # accuracy
        #print('train_precision: %f' % train_precision)
        #print('train_recall: %f' % train_recall) # uar (unweighted average recall)


    # ===============================
    # Validation
    # ===============================
    t_pred = model.predict(x_test)
    val_acc = np.mean(t_test == t_pred)
    
    # val_loss, val_acc, val_precision, val_recall = model.evaluate(x_test, t_test, metrics=['loss', 'accuracy','precision', 'recall'])
    if (verbose):
        print('val_acc: %f' % val_acc)
        #print('val_precision: %f' % val_precision)
        #print('val_recall: %f' % val_recall)
        print('train/fit time: {:.3f} ms'.format((time2-time1)/ (10 ** 6)))
    
    return train_acc, val_acc, (time2-time1)/ (10 ** 6)


def PerformenceTest(func, X, y, Ls = list(range(1, 80)), plot = False):
    '''
    Run one round performence test. 

    Parameters
    ----------
    func : a funciton callback that encapsulates a specific learning model.
    X,y : the dataset for supervised learning
    Ls: a list of hyper-parameter candidate values to test. The specific meaning of the hyper-parameter is determined by the func.
    plot: whether to display the generated curves.

    Returns
    -------
    Tacc : the training accuracies
    Vacc : the validation accuracies
    T : the training/fitting time
    '''

    Tacc = []
    Vacc = []
    T = []

    for L in Ls:
        ta, va, t = func(X, y, L)
        Tacc.append(ta)
        Vacc.append(va)
        T.append(t)

    if plot:        
        plt.figure(figsize = (12,4))
        plt.plot(Ls, Tacc, label="train acc")
        plt.plot(Ls, Vacc, label = "val acc")
        plt.legend()
        plt.show()

        plt.figure(figsize = (12,4))
        plt.plot(Ls, T, label="train/fit time (ms)")
        plt.legend()
        plt.show()
    
    return Tacc, Vacc, T


def PerformenceTests(func, X, y, Ls = list(range(1, 80)), N = 20):

    '''
    Run multiple rounds of performence test to get the average.

    Parameters
    ----------
    func : a funciton callback that encapsulates a specific learning model.
    X,y : the dataset for supervised learning
    Ls : a list of hyper-parameter candidate values to test. The specific meaning of the hyper-parameter is determined by the func.
    plot : whether to display the generated curves.
    N : rounds of test

    Returns
    -------
    MTacc : the averaged training accuracies
    MVacc : the averaged validation accuracies
    MT : the averaged training/fitting time
    '''


    Taccs = []
    Vaccs = []
    Ts = []
    
    for i in range(0, N):
        Tacc, Vacc, T = PerformenceTest(func, X, y, Ls = Ls, plot = False)
        Taccs.append(Tacc)
        Vaccs.append(Vacc)
        Ts.append(T)
        
    MTacc = np.array(Taccs).mean(axis = 0)
    MVacc = np.array(Vaccs).mean(axis = 0)
    MT = np.array(Ts).mean(axis = 0)
    
    plt.figure(figsize = (12,4))
    plt.plot(Ls, MTacc, label="train acc")
    plt.plot(Ls, MVacc, label = "val acc")
    plt.legend()
    plt.show()

    plt.figure(figsize = (12,4))
    plt.plot(Ls, MT, label="train/fit time (ms)")
    plt.legend()
    plt.show()
    
    return MTacc, MVacc, MT

#endregion

#region ensembles

def homo_stacking(X, y, create_base_estimator, Ns = [1, 2, 5], Ls = [1, 2, 5, 10, 20], repeat = 10, WITH_CONTEXT = True, xlabel = ''):
    '''
    create_base_estimator # a function that create base learner instanes
    Ns = [1, 2, 5] # number of base estimators
    Ls = [1, 2, 5, 10, 20] # base learner's specific hyper-parameter
    repeat = 10 # run multiple times to get the mean
    '''

    # pbar = tqdm(total = repeat * len(Ns) * len(Ls), position=0, leave=True) # stay on top

    plt.figure(figsize = (14, 6))

    for N in Ns:
        ACCs = []

        for L in Ls:        

            acc = 0

            for iter in range(repeat):

                X_train, X_test, y_train, y_test = train_test_split(
                    X, y
                )

                estimators = []
                for i in range(N):
                    estimators.append((str(i), create_base_estimator(L))) # relu is better than logistic

                clf = StackingClassifier(
                    estimators=estimators, final_estimator=LogisticRegression(), passthrough = WITH_CONTEXT
                )

                acc = acc + clf.fit(X_train, y_train).score(X_test, y_test)

                # pbar.update()

            ACCs.append(acc / repeat)

        plt.plot(Ls, ACCs, '--', label = 'N='+str(N), marker='o') # fillstyle='none'
        # plt.scatter(Ls, ACCs, s = 50)

    plt.gca().xaxis.set_major_locator(mticker.MultipleLocator(1)) # only show integer
    plt.legend()
    plt.xlabel('Hyper-parameter' + xlabel)
    plt.ylabel("Classfication Accuracy")
    plt.show()

    return ACCs
    
def hetero_stacking(X, y, estimators, repeat = 10, WITH_CONTEXT = True):
    '''
    estimators: a batch of base learners. Each learner should be derived from the BaseEstimator type.    
    '''
    acc = 0
    
    for iter in range(repeat):

        X_train, X_test, y_train, y_test = train_test_split(
            X, y
        )

        clf = StackingClassifier(
            estimators=estimators, final_estimator=LogisticRegression(), passthrough = WITH_CONTEXT
        )

        acc = acc + clf.fit(X_train, y_train).score(X_test, y_test)
    
    return acc / repeat

class FSSE(BaseEstimator, ClassifierMixin):
    '''
    Feature subspace based ensemble (FSSE)
    
    parameters
    ----------
    create_base_estimator_cv : { create_elmcv_instance, create_rvflcv_instance }. 
        a function callback that creates a cv optimized base estimator. 
    feature_split : an integer or 'all'. The interval or "window size" for each split. 
    meta_l1_ratios : l1_ratios to be tried for the meta-learner. We recommend > 0.5 values to achieve sparsity

    Sample Code
    -----------
    fsse = FSSE(create_elmcv_instance, feature_split = split)
    fsse.fit(X, y)
    acc = fsse.evaluate(X, y) # accuracy
    '''
    def __init__(self, create_base_estimator_cv, feature_split = 'all', meta_l1_ratios = [0.5,0.6,0.7,0.8,0.9,1.0]):

        self.create_base_estimator_cv = create_base_estimator_cv
        self.feature_split = feature_split
        self.meta_l1_ratios = meta_l1_ratios

    def fit(self, X, y):     
        
        if isinstance(self.feature_split, int):
            self.feature_split = abs(self.feature_split)

        if (self.feature_split == 'all' or self.feature_split >= X.shape[1]):
            self.feature_split = X.shape[1] 

        X_train, X_test, y_train, y_test = train_test_split(X, y)
        
        self.base_learners = {}
        predicts = []
        for i in range(0, X.shape[1], self.feature_split): # TODO: padding to N*feature_split, or special treatment for the last group.
            clfcv = self.create_base_estimator_cv().fit(X_train[:,i:i+ self.feature_split], y_train)
            self.base_learners[(i, i + self.feature_split)] = clfcv.clf.best_estimator_ # the inner best model
            predicts.append( clfcv.predict(X_test[:,i:i+ self.feature_split]) )

        predicts = np.array(predicts).T
        # print(predicts.shape)
        # print(X_test.shape, y_test.shape)
        self.meta_learner = LogisticRegressionCV(penalty = 'elasticnet',  # use elasticnet penalty to get sparse result
                                            l1_ratios = self.meta_l1_ratios, # require at least 0.5 L1 ratio for sparsity
                                           solver = 'saga' # only saga support elasticnet penalty
                                           ).fit(predicts ,y_test)  
        # print(meta_learner.l1_ratio_)
        # plt.plot(meta_learner.coef_[0])
        # return base_learners, meta_learner  

    def predict(self, X):
        predicts = []
        for b in self.base_learners:
            FS = X[:,b[0]:b[1]]
            yhat = self.base_learners[b].predict(FS)
            predicts.append(yhat)

        predicts = np.array(predicts).T
        return self.meta_learner.predict(predicts)
    
    def evaluate(self, X, y, metrics=['accuracy']):
        y_hat = self.predict(X)        
        return (y_hat == y).mean()

    # def plot_feature_importance():
    #    plot_feature_importance(np.abs(self.meta_learner.coef_[0]), 'FSSE FS Result') # need to include feature_importance.py

    def get_important_features(self):
        '''
        Sample Code
        -----------
        biggest_fsse_fs, fs_importance = fsse.get_important_features()
        xfsse = X_scaled[:,biggest_fsse_fs] # 前N个系数 non-zero
        plot_feature_importance(np.abs(fs_importance), 'FSSE FS Result') # require feature_importance.py

        return
        ------
        biggest_fsse_fs : the most important feature subset selected by meta-learner's coef
        fs_importance : the most important features' importance
        '''

        # print(fsse.meta_learner.l1_ratio_)
        N = np.count_nonzero(self.meta_learner.coef_[0])
        
        biggest_fsse_gs = (np.argsort(np.abs(self.meta_learner.coef_[0]))[-N:])[::-1] # take last N item indices and reverse (ord desc)
        biggest_fsse_fs = []
        ranges = list(self.base_learners)

        for idx in biggest_fsse_gs:
            biggest_fsse_fs = biggest_fsse_fs + list(  range(ranges[idx][0], ranges[idx][1]) ) # list(range(idx*self.feature_split, (idx+1)*self.feature_split))
    
        biggest_fsse_fs = np.array(biggest_fsse_fs)

        fs_importance = []

        for w in self.meta_learner.coef_[0]:
            fs_importance = fs_importance + [w]*self.feature_split
    
        fs_importance = np.array(fs_importance)

        return biggest_fsse_fs, fs_importance

def fsse_homo_stacking(X, y, create_base_estimator_cv, split_range = range(1, 20), repeat = 3, summary = 'median'):

    accs = []
    r = split_range
    N = repeat

    for split in r:

        for i in range(N):

            accs_repeat = []

            fsse = FSSE(create_base_estimator_cv, feature_split = split)
            fsse.fit(X, y)
            accs_repeat.append(fsse.evaluate(X, y))

        if summary == 'median':
            accs.append( np.median(accs_repeat) )
        else:
            accs.append( np.mean(accs_repeat) ) # otherwise, mean

    plt.figure(figsize = (12,3))
    plt.scatter(r,accs)
    plt.plot(r, accs)
    plt.show()
    
    return accs

#endregion



import numpy as np
from sklearn.neural_network import BernoulliRBM

class RBM:
  
  def __init__(self, num_visible, num_hidden):
    self.num_hidden = num_hidden
    self.num_visible = num_visible
    self.debug_print = True

    # Initialize a weight matrix, of dimensions (num_visible x num_hidden), using
    # a uniform distribution between -sqrt(6. / (num_hidden + num_visible))
    # and sqrt(6. / (num_hidden + num_visible)). One could vary the 
    # standard deviation by multiplying the interval with appropriate value.
    # Here we initialize the weights with mean 0 and standard deviation 0.1. 
    # Reference: Understanding the difficulty of training deep feedforward 
    # neural networks by Xavier Glorot and Yoshua Bengio
    np_rng = np.random.RandomState(1234)

    self.weights = np.asarray(np_rng.uniform(
			low=-0.1 * np.sqrt(6. / (num_hidden + num_visible)),
                       	high=0.1 * np.sqrt(6. / (num_hidden + num_visible)),
                       	size=(num_visible, num_hidden)))


    # Insert weights for the bias units into the first row and first column.
    self.weights = np.insert(self.weights, 0, 0, axis = 0)
    self.weights = np.insert(self.weights, 0, 0, axis = 1)

  def train(self, data, max_epochs = 1000, learning_rate = 0.1):
    """
    Train the machine.

    Parameters
    ----------
    data: A matrix where each row is a training example consisting of the states of visible units.    
    """

    num_examples = data.shape[0]

    # Insert bias units of 1 into the first column.
    data = np.insert(data, 0, 1, axis = 1)

    for epoch in range(max_epochs):      
      # Clamp to the data and sample from the hidden units. 
      # (This is the "positive CD phase", aka the reality phase.)
      pos_hidden_activations = np.dot(data, self.weights)      
      pos_hidden_probs = self._logistic(pos_hidden_activations)
      pos_hidden_probs[:,0] = 1 # Fix the bias unit.
      pos_hidden_states = pos_hidden_probs > np.random.rand(num_examples, self.num_hidden + 1)
      # Note that we're using the activation *probabilities* of the hidden states, not the hidden states       
      # themselves, when computing associations. We could also use the states; see section 3 of Hinton's 
      # "A Practical Guide to Training Restricted Boltzmann Machines" for more.
      pos_associations = np.dot(data.T, pos_hidden_probs)

      # Reconstruct the visible units and sample again from the hidden units.
      # (This is the "negative CD phase", aka the daydreaming phase.)
      neg_visible_activations = np.dot(pos_hidden_states, self.weights.T)
      neg_visible_probs = self._logistic(neg_visible_activations)
      neg_visible_probs[:,0] = 1 # Fix the bias unit.
      neg_hidden_activations = np.dot(neg_visible_probs, self.weights)
      neg_hidden_probs = self._logistic(neg_hidden_activations)
      # Note, again, that we're using the activation *probabilities* when computing associations, not the states 
      # themselves.
      neg_associations = np.dot(neg_visible_probs.T, neg_hidden_probs)

      # Update weights.
      self.weights += learning_rate * ((pos_associations - neg_associations) / num_examples)

      error = np.sum((data - neg_visible_probs) ** 2)
      if self.debug_print:
        print("Epoch %s: error is %s" % (epoch, error))

  def run_visible(self, data):
    """
    Assuming the RBM has been trained (so that weights for the network have been learned),
    run the network on a set of visible units, to get a sample of the hidden units.
    
    Parameters
    ----------
    data: A matrix where each row consists of the states of the visible units.
    
    Returns
    -------
    hidden_states: A matrix where each row consists of the hidden units activated from the visible
    units in the data matrix passed in.
    """
    
    num_examples = data.shape[0]
    
    # Create a matrix, where each row is to be the hidden units (plus a bias unit)
    # sampled from a training example.
    hidden_states = np.ones((num_examples, self.num_hidden + 1))
    
    # Insert bias units of 1 into the first column of data.
    data = np.insert(data, 0, 1, axis = 1)

    # Calculate the activations of the hidden units.
    hidden_activations = np.dot(data, self.weights)
    # Calculate the probabilities of turning the hidden units on.
    hidden_probs = self._logistic(hidden_activations)
    # Turn the hidden units on with their specified probabilities.
    hidden_states[:,:] = hidden_probs > np.random.rand(num_examples, self.num_hidden + 1)
    # Always fix the bias unit to 1.
    # hidden_states[:,0] = 1
  
    # Ignore the bias units.
    hidden_states = hidden_states[:,1:]
    return hidden_states
    
  # TODO: Remove the code duplication between this method and `run_visible`?
  def run_hidden(self, data):
    """
    Assuming the RBM has been trained (so that weights for the network have been learned),
    run the network on a set of hidden units, to get a sample of the visible units.

    Parameters
    ----------
    data: A matrix where each row consists of the states of the hidden units.

    Returns
    -------
    visible_states: A matrix where each row consists of the visible units activated from the hidden
    units in the data matrix passed in.
    """

    num_examples = data.shape[0]

    # Create a matrix, where each row is to be the visible units (plus a bias unit)
    # sampled from a training example.
    visible_states = np.ones((num_examples, self.num_visible + 1))

    # Insert bias units of 1 into the first column of data.
    data = np.insert(data, 0, 1, axis = 1)

    # Calculate the activations of the visible units.
    visible_activations = np.dot(data, self.weights.T)
    # Calculate the probabilities of turning the visible units on.
    visible_probs = self._logistic(visible_activations)
    # Turn the visible units on with their specified probabilities.
    visible_states[:,:] = visible_probs > np.random.rand(num_examples, self.num_visible + 1)
    # Always fix the bias unit to 1.
    # visible_states[:,0] = 1

    # Ignore the bias units.
    visible_states = visible_states[:,1:]
    return visible_states
    
  def daydream(self, num_samples):
    """
    Randomly initialize the visible units once, and start running alternating Gibbs sampling steps
    (where each step consists of updating all the hidden units, and then updating all of the visible units),
    taking a sample of the visible units at each step.
    Note that we only initialize the network *once*, so these samples are correlated.

    Returns
    -------
    samples: A matrix, where each row is a sample of the visible units produced while the network was
    daydreaming.
    """

    # Create a matrix, where each row is to be a sample of of the visible units 
    # (with an extra bias unit), initialized to all ones.
    samples = np.ones((num_samples, self.num_visible + 1))

    # Take the first sample from a uniform distribution.
    samples[0,1:] = np.random.rand(self.num_visible)

    # Start the alternating Gibbs sampling.
    # Note that we keep the hidden units binary states, but leave the
    # visible units as real probabilities. See section 3 of Hinton's
    # "A Practical Guide to Training Restricted Boltzmann Machines"
    # for more on why.
    for i in range(1, num_samples):
      visible = samples[i-1,:]

      # Calculate the activations of the hidden units.
      hidden_activations = np.dot(visible, self.weights)      
      # Calculate the probabilities of turning the hidden units on.
      hidden_probs = self._logistic(hidden_activations)
      # Turn the hidden units on with their specified probabilities.
      hidden_states = hidden_probs > np.random.rand(self.num_hidden + 1)
      # Always fix the bias unit to 1.
      hidden_states[0] = 1

      # Recalculate the probabilities that the visible units are on.
      visible_activations = np.dot(hidden_states, self.weights.T)
      visible_probs = self._logistic(visible_activations)
      visible_states = visible_probs > np.random.rand(self.num_visible + 1)
      samples[i,:] = visible_states

    # Ignore the bias units (the first column), since they're always set to 1.
    return samples[:,1:]        
      
  def _logistic(self, x):
    return 1.0 / (1 + np.exp(-x))

if __name__ == '__main__':
  r = RBM(num_visible = 6, num_hidden = 2)
  training_data = np.array([[1,1,1,0,0,0],[1,0,1,0,0,0],[1,1,1,0,0,0],[0,0,1,1,1,0], [0,0,1,1,0,0],[0,0,1,1,1,0]])
  r.train(training_data, max_epochs = 5000)
  print(r.weights)
  user = np.array([[0,0,0,1,1,0]])
  print(r.run_visible(user))

